#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/28 00:01
@Author  : alexanderwu
@File    : base_store.py
"""
from abc import ABC, abstractmethod
from pathlib import Path


class BaseStore(ABC):
    """FIXME: consider add_index, set_index and think 颗粒度"""

    @abstractmethod
    def search(self, query, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def write(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def add(self, *args, **kwargs):
        raise NotImplementedError


class LocalStore(BaseStore, ABC):
    def __init__(self, raw_data: Path, cache_dir: Path = None):
        if not raw_data:
            raise FileNotFoundError
        self.raw_data = raw_data
        if not cache_dir:
            cache_dir = raw_data.parent
        self.cache_dir = cache_dir
        self.store = self._load()
        if not self.store:
            self.store = self.write()

    def _get_index_and_store_fname(self):
        fname = self.raw_data.name.split('.')[0]
        index_file = self.cache_dir / f"{fname}.index"
        store_file = self.cache_dir / f"{fname}.pkl"
        return index_file, store_file

    @abstractmethod
    def _load(self):
        raise NotImplementedError

    @abstractmethod
    def _write(self, docs, metadatas):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/11 16:15
@Author  : alexanderwu
@File    : singleton.py
"""
import abc


class Singleton(abc.ABCMeta, type):
    """
    Singleton metaclass for ensuring only one instance of a class.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Call method for the singleton metaclass."""
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/5 23:08
@Author  : alexanderwu
@File    : openai.py
"""
import asyncio
import time
import random
import string
from functools import wraps
from typing import NamedTuple, Union, List, Dict, Any

import httpx
import openai
from openai import AsyncOpenAI, OpenAI

from schema_agents.config import CONFIG
from schema_agents.logs import logger
from schema_agents.provider.base_gpt_api import BaseGPTAPI
from schema_agents.utils.singleton import Singleton
from schema_agents.utils.token_counter import (
    TOKEN_COSTS,
    count_message_tokens,
    count_string_tokens,
    get_max_completion_tokens,
)
from schema_agents.utils.common import EventBus, current_session
from contextvars import copy_context

def retry(max_retries):
    def decorator(f):
        @wraps(f)
        async def wrapper(*args, **kwargs):
            for i in range(max_retries):
                try:
                    return await f(*args, **kwargs)
                except Exception:
                    if i == max_retries - 1:
                        raise
                    await asyncio.sleep(2 ** i)
        return wrapper
    return decorator


class RateLimiter:
    """Rate control class, each call goes through wait_if_needed, sleep if rate control is needed"""
    def __init__(self, rpm):
        self.last_call_time = 0
        self.interval = 1.1 * 60 / rpm  # Here 1.1 is used because even if the calls are made strictly according to time, they will still be QOS'd; consider switching to simple error retry later
        self.rpm = rpm

    def split_batches(self, batch):
        return [batch[i:i + self.rpm] for i in range(0, len(batch), self.rpm)]

    async def wait_if_needed(self, num_requests):
        current_time = time.time()
        elapsed_time = current_time - self.last_call_time

        if elapsed_time < self.interval * num_requests:
            remaining_time = self.interval * num_requests - elapsed_time
            logger.info(f"sleep {remaining_time}")
            await asyncio.sleep(remaining_time)

        self.last_call_time = time.time()


class Costs(NamedTuple):
    total_prompt_tokens: int
    total_completion_tokens: int
    total_cost: float
    total_budget: float


class CostManager(metaclass=Singleton):
    """计算使用接口的开销"""
    def __init__(self):
        self.total_prompt_tokens = 0
        self.total_completion_tokens = 0
        self.total_cost = 0
        self.total_budget = 0

    def update_cost(self, prompt_tokens, completion_tokens, model):
        """
        Update the total cost, prompt tokens, and completion tokens.

        Args:
        prompt_tokens (int): The number of tokens used in the prompt.
        completion_tokens (int): The number of tokens used in the completion.
        model (str): The model used for the API call.
        """
        self.total_prompt_tokens += prompt_tokens
        self.total_completion_tokens += completion_tokens
        cost = (
            prompt_tokens * TOKEN_COSTS[model]["prompt"]
            + completion_tokens * TOKEN_COSTS[model]["completion"]
        ) / 1000
        self.total_cost += cost
        logger.info(f"Total running cost: ${self.total_cost:.3f} | Max budget: ${CONFIG.max_budget:.3f} | "
                    f"Current cost: ${cost:.3f}, {prompt_tokens=}, {completion_tokens=}")
        CONFIG.total_cost = self.total_cost

    def get_total_prompt_tokens(self):
        """
        Get the total number of prompt tokens.

        Returns:
        int: The total number of prompt tokens.
        """
        return self.total_prompt_tokens

    def get_total_completion_tokens(self):
        """
        Get the total number of completion tokens.

        Returns:
        int: The total number of completion tokens.
        """
        return self.total_completion_tokens

    def get_total_cost(self):
        """
        Get the total cost of API calls.

        Returns:
        float: The total cost of API calls.
        """
        return self.total_cost

    def get_costs(self) -> Costs:
        """获得所有开销"""
        return Costs(self.total_prompt_tokens, self.total_completion_tokens, self.total_cost, self.total_budget)


class OpenAIGPTAPI(BaseGPTAPI, RateLimiter):
    """
    Check https://platform.openai.com/examples for examples
    """
    def __init__(self, model=None, seed=None, temperature=None):
        self.__init_openai(CONFIG)
        self.model = model or CONFIG.openai_api_model
        self.temperature = temperature or CONFIG.openai_temperature
        self.seed = seed or CONFIG.openai_seed
        self.auto_max_tokens = False
        self._cost_manager = CostManager()
        RateLimiter.__init__(self, rpm=self.rpm)
        logger.info(f"OpenAI API model: {self.model}")

    def __init_openai(self, config):
        if config.openai_proxy:
            openai.http_client = httpx.Client(
                proxies=config.openai_proxy,
                timeout=20,
            )
        if config.openai_api_type:
            openai.api_type = config.openai_api_type
            openai.api_version = config.openai_api_version
        self.aclient = AsyncOpenAI(
            api_key=config.openai_api_key,
            base_url=config.openai_api_base,
        )
        self.client = OpenAI(
            api_key=config.openai_api_key,
            base_url=config.openai_api_base,
        )
        self.rpm = int(config.get("RPM", 10))

    async def _achat_completion_stream(self, messages: list[dict], event_bus: EventBus=None, **kwargs) -> str:
        response = await self.aclient.chat.completions.create(
            **self._cons_kwargs(messages, functions=kwargs.get("functions")),
            stream=True,
            **kwargs
        )

        # create variables to collect the stream of chunks
        collected_chunks = []
        collected_messages = []
        function_call_detected = False
        func_call = {}
        query_id = ''.join(random.choices(string.ascii_lowercase + string.digits, k=8))
        session_id = current_session.get() if current_session in copy_context() else None
        # iterate through the stream of events
        async for chunk in response:
            chunk = chunk.dict()
            collected_chunks.append(chunk)  # save the event response
            chunk_message = chunk['choices'][0]['delta']  # extract the message
            collected_messages.append(chunk_message)  # save the message
            
            if "function_call" in chunk_message and chunk_message["function_call"]:
                if "name" in chunk_message["function_call"] and chunk_message["function_call"]["name"]:
                    func_call["name"] = chunk_message["function_call"]["name"]
                    if event_bus:
                        event_bus.emit("stream", {"query_id": query_id, "session_id": session_id, "type": "function_call", "name": func_call["name"], "arguments": func_call.get("arguments", ""), "status": "start"})
                    # if not function_call_detected:
                    #     print(func_call["name"] + "(", end="")
                if "arguments" in chunk_message["function_call"]:
                    if "arguments" not in func_call:
                        func_call["arguments"] = ""
                    func_call["arguments"] += chunk_message["function_call"]["arguments"]
                    # print(chunk_message["function_call"]["arguments"], end="")
                function_call_detected = True
            if event_bus:
                if function_call_detected:
                    if chunk["choices"][0].get("finish_reason") in ["function_call", "stop"]:
                        event_bus.emit("function_call", func_call)
                        event_bus.emit("stream", {"query_id": query_id, "session_id": session_id, "type": "function_call", "name": func_call["name"], "arguments": func_call["arguments"], "status": "finished"})
                    elif chunk_message["function_call"].get("arguments"):
                        event_bus.emit("stream", {"query_id": query_id, "session_id": session_id, "type": "function_call", "name": func_call["name"], "arguments": chunk_message["function_call"]["arguments"], "status": "in_progress"})
                elif "content" in chunk_message and chunk_message["content"]:
                    event_bus.emit("stream", {"query_id": query_id, "session_id": session_id, "type": "text", "content": chunk_message["content"]})

        # if function_call_detected:
        #     print(")", end="")
        # print()
        
        if function_call_detected:
            full_reply_content = func_call
            # TODO: check if the usage calculation is correct
            usage = self._calc_usage(messages, f"{func_call['name']}({func_call['arguments']})", functions=kwargs.get("functions", None))
        else:
            full_reply_content = ''.join([m.get('content', '') for m in collected_messages])
            usage = self._calc_usage(messages, full_reply_content, functions=kwargs.get("functions", None))
        self._update_costs(usage)
        if event_bus:
            event_bus.emit("completion", full_reply_content)
        return full_reply_content


    def _cons_kwargs(self, messages: list[dict], functions: list[dict]=None) -> dict:
        kwargs = {
            "messages": messages,
            "max_tokens": self.get_max_tokens(messages, functions=functions),
            "n": 1,
            "stop": None,
            "temperature": self.temperature,
            "timeout": 3,
            "seed": self.seed,
        }
        if CONFIG.openai_api_type == "azure":
            if CONFIG.deployment_name and CONFIG.deployment_id:
                raise ValueError("You can only use one of the `deployment_id` or `deployment_name` model")
            elif not CONFIG.deployment_name and not CONFIG.deployment_id:
                raise ValueError("You must specify `DEPLOYMENT_NAME` or `DEPLOYMENT_ID` parameter")
            kwargs_mode = (
                {"engine": CONFIG.deployment_name}
                if CONFIG.deployment_name
                else {"deployment_id": CONFIG.deployment_id}
            )
        else:
            kwargs_mode = {"model": self.model}
        kwargs.update(kwargs_mode)
        return kwargs

    async def _achat_completion(self, messages: list[dict], event_bus: EventBus=None, **kwargs) -> dict:
        kwargs.update(self._cons_kwargs(messages, functions=kwargs.get("functions")))
        rsp = await self.aclient.chat.completions.create(**kwargs)
        self._update_costs(rsp.get('usage'))
        if event_bus:
            event_bus.emit("completion", rsp)
        return rsp

    def _chat_completion(self, messages: list[dict], event_bus: EventBus=None) -> dict:
        rsp = self.client.chat.completions.create(**self._cons_kwargs(messages))
        self._update_costs(rsp)
        if event_bus:
            event_bus.emit("completion", rsp)
        return rsp

    def completion(self, messages: list[dict], event_bus: EventBus=None) -> dict:
        # if isinstance(messages[0], Message):
        #     messages = self.messages_to_dict(messages)
        rsp = self._chat_completion(messages, event_bus=event_bus)
        return rsp

    async def acompletion(self, messages: list[dict], event_bus: EventBus=None) -> dict:
        # if isinstance(messages[0], Message):
        #     messages = self.messages_to_dict(messages)
        rsp = await self._achat_completion_stream(messages, event_bus=event_bus)
        return rsp
    
    async def acompletion_function(self, messages: list[dict], functions: List[Dict[str, Any]]=None, function_call: Union[str, Dict[str, str]]=None, event_bus: EventBus=None) -> dict:
        # if isinstance(messages[0], Message):
        #     messages = self.messages_to_dict(messages)
        rsp = await self._achat_completion_stream(messages, functions=functions, function_call=function_call, event_bus=event_bus)
        return rsp

    @retry(max_retries=6)
    async def acompletion_text(self, messages: list[dict], stream=False, event_bus: EventBus=None) -> str:
        """when streaming, print each token in place."""
        if stream:
            return await self._achat_completion_stream(messages, event_bus=event_bus)
        rsp = await self._achat_completion(messages, event_bus=event_bus)
        return self.get_choice_text(rsp)

    def _calc_usage(self, messages: list[dict], rsp: str, functions: List[dict]=None) -> dict:
        usage = {}
        prompt_tokens = count_message_tokens(messages, self.model, functions=functions)
        completion_tokens = count_string_tokens(rsp, self.model)
        usage['prompt_tokens'] = prompt_tokens
        usage['completion_tokens'] = completion_tokens
        return usage

    async def acompletion_batch(self, batch: list[list[dict]], event_bus: EventBus=None) -> list[dict]:
        """返回完整JSON"""
        split_batches = self.split_batches(batch)
        all_results = []

        for small_batch in split_batches:
            logger.info(small_batch)
            await self.wait_if_needed(len(small_batch))

            future = [self.acompletion(prompt, event_bus=event_bus) for prompt in small_batch]
            results = await asyncio.gather(*future)
            logger.info(results)
            all_results.extend(results)

        return all_results

    async def acompletion_batch_text(self, batch: list[list[dict]], event_bus: EventBus=None) -> list[str]:
        """仅返回纯文本"""
        raw_results = await self.acompletion_batch(batch, event_bus=event_bus)
        results = []
        for idx, raw_result in enumerate(raw_results, start=1):
            result = self.get_choice_text(raw_result)
            results.append(result)
            logger.info(f"Result of task {idx}: {result}")
        return results

    def _update_costs(self, usage: dict):
        prompt_tokens = int(usage['prompt_tokens'])
        completion_tokens = int(usage['completion_tokens'])
        self._cost_manager.update_cost(prompt_tokens, completion_tokens, self.model)

    def get_costs(self) -> Costs:
        return self._cost_manager.get_costs()

    def get_max_tokens(self, messages: list[dict], functions: List[dict]=None):
        if not self.auto_max_tokens:
            return CONFIG.max_tokens_rsp
        return get_max_completion_tokens(messages, functions, self.model, CONFIG.max_tokens_rsp)

    def moderation(self, content: Union[str, list[str]]):
        try:
            if not content:
                logger.error("content cannot be empty!")
            else:
                rsp = self._moderation(content=content)
                return rsp
        except Exception as e:
            logger.error(f"moderating failed:{e}")

    def _moderation(self, content: Union[str, list[str]]):
        rsp = self.client.chat.moderations.create(input=content)
        return rsp

    async def amoderation(self, content: Union[str, list[str]]):
        try:
            if not content:
                logger.error("content cannot be empty!")
            else:
                rsp = await self._amoderation(content=content)
                return rsp
        except Exception as e:
            logger.error(f"moderating failed:{e}")

    async def _amoderation(self, content: Union[str, list[str]]):
        rsp = await self.aclient.chat.moderations.create(input=content)
        return rsp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
from typing import List, Optional, Union
from pydantic import BaseModel, Field
from schema_agents.teams import Team
from schema_agents.role import Role
from schema_agents.schema import Message

class SoftwareRequirementDocument(BaseModel):
    """Write Software Requirement Document."""
    product_goals: List[str] = Field(description="Up to 3 clear, orthogonal product goals. If the requirement itself is simple, the goal should also be simple")

class AnalysisScript(BaseModel):
    """Python script for image analysis."""
    script: str = Field(description="Python script for image analysis")

@pytest.mark.asyncio
async def test_team():
    async def create_user_requirements(query: str, role: Role) -> SoftwareRequirementDocument:
        """Create user requirements."""
        response = await role.aask(query, SoftwareRequirementDocument)
        return response
        
    bioimage_analyst = Role(name="Alice",
                profile="BioImage Analyst",
                goal="Efficiently communicate with the user and translate the user's needs into software requirements",
                constraints=None,
                actions=[create_user_requirements])
    
    async def write_code(req: SoftwareRequirementDocument, role: Role) -> AnalysisScript:
        """Write code for image analysis."""
        code = await role.aask(req, AnalysisScript)
        return code
    
    coder = Role(name="Bob",
                profile="Coder",
                goal="Write code for image analysis.",
                constraints=None,
                actions=[write_code])
    
    team = Team(name="Test Team")
    team.hire([bioimage_analyst, coder])
    event_bus = team.get_event_bus()
    event_bus.register_default_events()
    responses = await team.handle(Message(role="Bot", content="Create a segmentation software"))
    assert isinstance(responses[1].data, SoftwareRequirementDocument)
    assert isinstance(responses[2].data, AnalysisScript)
    
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/8 22:12
@Author  : alexanderwu
@File    : schema.py
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TypedDict, Callable

from pydantic import BaseModel

from schema_agents.logs import logger


class RawMessage(TypedDict):
    content: str
    role: str


@dataclass
class Message:
    """list[<role>: <content>]"""
    content: str
    data: BaseModel = field(default=None)
    role: str = field(default='user')  # system / user / assistant
    cause_by: Callable = field(default=None)
    processed_by: set['Role'] = field(default_factory=set)
    session_ids: list[str] = field(default_factory=list)

    def __str__(self):
        return f"{self.role}: {self.content}"

    def __repr__(self):
        return self.__str__()

    def to_dict(self) -> dict:
        return {
            "role": self.role,
            "content": self.content
        }


@dataclass
class UserMessage(Message):
    """便于支持OpenAI的消息"""
    def __init__(self, content: str):
        super().__init__(content, 'user')


@dataclass
class SystemMessage(Message):
    """便于支持OpenAI的消息"""
    def __init__(self, content: str):
        super().__init__(content, 'system')


@dataclass
class AIMessage(Message):
    """便于支持OpenAI的消息"""
    def __init__(self, content: str):
        super().__init__(content, 'assistant')


@dataclass
class MemoryChunk:
    """list[<role>: <content>]"""
    index: str
    content: BaseModel = field(default=None)
    category: str = field(default=None)

    def __str__(self):
        return f"memory chunk: {self.index}"

    def __repr__(self):
        return self.__str__()

    def to_dict(self) -> dict:
        return {
            "index": self.index
        }


if __name__ == '__main__':
    test_content = 'test_message'
    msgs = [
        UserMessage(test_content),
        SystemMessage(test_content),
        AIMessage(test_content),
        Message(test_content, role='QA')
    ]
    logger.info(msgs)



from openai import OpenAI

client = OpenAI()

response = client.chat.completions.create(
    model="gpt-4-vision-preview",
    messages=[
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "What’s in this image?"},
                {
                    "type": "image_url",
                    "image_url": {
                        "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",
                    },
                },
            ],
        }
    ],
    max_tokens=300,
)

print(response.choices[0])


from typing import List, Optional

from pydantic import BaseModel, Field
from enum import Enum

# Common description for Python function scripts
common_desc = {
    "script": "Python script that defines functions according to the requirements. Includes imports, function definition, logic, and implementation.",
    "names": "List of function names.",
    "pip_packages": "Required Python pip packages. Reuse existing libraries and prioritize common libraries.",
    "test_script": "Script for testing the Python function. Includes test cases, validation logic, and assertions.",
    "docstring": "Brief notes for usage, debugging, potential error fixing, and further improvements."
}


class GenerateUserForm(BaseModel):
    """Based on user query, create a JSON Schema Form Dialog using react-jsonschema-form to get more information about what the user want to create.
    The aim is to gather the information needed to create the software/tool requirement document or perform data acquisition.
    Whenever possible, try to propose the options for the user to choose from, instead of asking the user to type in the text."""
    form_schema: str = Field(description="json schema for the fields, in yaml format")
    ui_schema: Optional[str] = Field(None, description="customized ui schema for rendering the form, json string, no need to escape quotes, in yaml format")
    submit_label: Optional[str] = Field("Submit", description="Submit button label")

class UserRequirements(BaseModel):
    """User requirements for developing a software."""
    goal: str = Field(description="The goal of the user.")
    ui: str = Field(description="Requirements for the user interface.")
    data: str = Field(description="Requirements for the data.")
    validation: str = Field(description="Additional information for testing, e.g. test data and the expected outcome")
    notes: str = Field(description="Additional notes.")

class UserClarification(BaseModel):
    """User submitted form data."""
    user_query: str = Field(description="The original user query")
    form_data: str = Field(description="Form data in json format")

class GetExtraInformation(BaseModel):
    """Extra information needed to be able to work on the task."""
    content: str = Field(description="The information.")
    summary: str = Field(description="Summary of what you already get.")


class FunctionMemory(BaseModel):
    """Functions to be saved in the long term memory."""
    function_name: str = Field(default="", description="Function name")
    code: str = Field(default="", description="original code of the function")
    lang: str = Field(default="", description="function language")
    args: List[str] = Field(default=[], description="arguments of the function")
    
class ErrorMemory(BaseModel):
    """Experience of making errors to be saved in the long term memory."""    
    error: str = Field(default="", description="Error description")
    cause_by: str = Field(default="", description="Cause of the error")
    solution: str = Field(default="", description="Solution to fix the error")

class ExperienceMemory(BaseModel):
    """Experience to be saved in the long term memory."""    
    summary: str = Field(default="", description="Summary of the experience. This will be used as index in the long term memory for retrival.")
    keypoints: Optional[str] = Field(default="", description="Key points to remember for this experience.")

CODING_RULES = """
Important Rules for Coding:
- Use `window.python` to refer the external python functions
- Use tailwindcss for styling (the page has `https://cdn.tailwindcss.com` loaded)
- DO NOT user other libraries besides React and React DOM
"""

class ReactUI(BaseModel):
    """Defines the ImJoy UI plugin using React."""
    id: str = Field(..., description="a short id of the application")
    root_element: str = Field("root", description="The root element ID where the React app will be attached.")
    react_version: str = Field("17.0.2", description="Version of React to use.")
    react_dom_version: str = Field("17.0.2", description="Version of ReactDOM to use.")
    babel_version: str = Field("6.26.0", description="Version of Babel to transpile JSX.")
    jsx_script: str = Field(..., description="JSX script defining the React app. It will be transpiled with Babel and the dom should be mounted to the root element. DO NOT use import statements." + CODING_RULES) # In the script it must export a set of function as the ImJoy plugin API. e.g.: `api.export({setup, run, show_image: showImage, etc....})
    # test_script: str = Field(..., description="Test script for calling the exported ImJoy API for testing. Should include test cases, expected outcomes, and validation logic.")

class PythonFunctionScript(BaseModel):
    """Represents a Python function and test script with all its properties."""
    function_names: List[str] = Field(..., description=common_desc['names'])
    function_script: str = Field(..., description=common_desc['script'])
    pip_packages: List[str] = Field(..., description=common_desc['pip_packages'])
    test_script: str = Field(..., description=common_desc['test_script'])
    docstring: Optional[str] = Field(None, description=common_desc['docstring'])
   

class PythonFunctionScriptWithLineNumber(BaseModel):
    """Python function and test script with line number for display and editing"""
    function_script_with_line_number: str = Field(..., description=common_desc['script'])
    pip_packages: List[str] = Field(..., description=common_desc['pip_packages'])
    test_script_with_line_number: str = Field(..., description=common_desc['test_script'])
    docstring: Optional[str] = Field(None, description=common_desc['docstring'])

# Enum for operations
class OperationEnum(str, Enum):
    ADD = "add"
    REMOVE = "remove"
    REPLACE = "replace"


# Description for Change operations
change_desc = {
    "operation": "The operation type ('add', 'remove', 'replace')",
    "line": "The line number (of the original script) affected by the change",
    "content": "The content to be added or replaced"
}

# BaseModel for changes in Python function scripts
class Change(BaseModel):
    operation: OperationEnum = Field(..., description=change_desc['operation'])
    line: int = Field(..., description=change_desc['line'])
    content: str = Field(None, description=change_desc['content'])

class PythonFunctionScriptChanges(BaseModel):
    """Changes to the Python function and test script."""
    explanation: str = Field(..., description="Brief explanation on why the changes is needed.")
    function_script_changes: Optional[List[Change]] = Field(None, description="List of changes to the function script.")
    updated_pip_packages: List[str] = Field(..., description="Updated pip packages.")
    test_script_changes: Optional[List[Change]] = Field(None, description="List of changes to the test script.")
    experience: Optional[ExperienceMemory] = Field(None, description="Experience for successfully fixing the errors, this will be noted down in the long term memory to avoid making the same mistake again.")

class PythonFunctionRequirement(BaseModel):
    """Python Function Requirement
    Providing detailed information for implementing the python function."""
    function_name: str = Field(..., description="Name of the python function to be implemented.")
    function_signature: str = Field(..., description="Signature of the python function, outlining the expected parameters and return type.")
    feature_instructions: str = Field(..., description="Clear instructions on how to implement the feature, including code snippets, logic, and algorithms if necessary.")
    context_requirements: Optional[str] = Field(default=None, description="Additional context or prerequisites required for the python function, including dependencies on other functions, modules, or data.")
    testing_requirements: Optional[str] = Field(default=None, description="Instructions or guidelines on how to test the python function, including expected outcomes, edge cases, and validation criteria.")


class ReactUIRequirement(BaseModel):
    """React UI Requirement
    The aim is to create a web UI for the main script in python to obtain user input, display results, and allow for interaction. 
    The exported imjoy plugin api function will be called inside the main function to interact with the user."""
    plugin_name: str = Field(..., description="Name of the React plugin for main function referencing.")
    ui_layout: str = Field(..., description="Description of the UI layout, including positioning of elements.")
    interaction_patterns: str = Field(..., description="Details of interaction patterns, such as clicks, swipes, etc.")
    functionalities: str = Field(..., description="Details of the functions that need to be implemented, such as selecting an image, segmenting cells, etc.")
    user_flows: str = Field(..., description="Outline of the user flow, describing how the user moves through the application.")
    # imjoy_plugin_api: List[ReactUIApiFunction] = Field(..., description="List of functions for the main function to call. These functions are used to configure the UI, display results, register callbacks, etc.")

class SoftwareRequirement(BaseModel):
    """Software Requirement
    The the software requirement is to used to instruct the developers to create a set of python functions with web UI built with react.js according to user's request.
    """
    id: str = Field(..., description="a short id of the application")
    original_requirements: str = Field(..., description="The polished complete original requirements from the user.")
    python_function_requirements: Optional[list[PythonFunctionRequirement]] = Field(..., description="A list of requirements for the python functions which will be called in the web UI.")
    react_ui_requirements: Optional[ReactUIRequirement] = Field(description="User interface requirements for the react.js web UI. The UI will be used to interact with the user and call the python functions (made available under a global variable named `pythonFunctions` using imjoy-rpc). E.g. `pythonFunctions.load_data(...)` can be called in a button click callback.")
    additional_notes: str = Field(default="", description="Any additional notes or requirements that need to be considered.")





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/25 10:20
@Author  : alexanderwu
@File    : faiss_store.py
"""
import pickle
from pathlib import Path
from typing import Optional

import faiss
from langchain.embeddings import OpenAIEmbeddings
from langchain.vectorstores import FAISS

from schema_agents.memory.base_store import LocalStore
from schema_agents.logs import logger


class FaissStore(LocalStore):
    def __init__(self, raw_data: Path, cache_dir=None, meta_col='source', content_col='output'):
        self.meta_col = meta_col
        self.content_col = content_col
        super().__init__(raw_data, cache_dir)

    def _load(self) -> Optional["FaissStore"]:
        index_file, store_file = self._get_index_and_store_fname()
        if not (index_file.exists() and store_file.exists()):
            logger.info("Missing at least one of index_file/store_file, load failed and return None")
            return None
        index = faiss.read_index(str(index_file))
        with open(str(store_file), "rb") as f:
            store = pickle.load(f)
        store.index = index
        return store

    def _write(self, docs, metadatas):
        store = FAISS.from_texts(docs, OpenAIEmbeddings(openai_api_version="2020-11-07"), metadatas=metadatas)
        return store

    def persist(self):
        index_file, store_file = self._get_index_and_store_fname()
        store = self.store
        index = self.store.index
        faiss.write_index(store.index, str(index_file))
        store.index = None
        with open(store_file, "wb") as f:
            pickle.dump(store, f)
        store.index = index

    def search(self, query, expand_cols=False, sep='\n', *args, k=5, **kwargs):
        rsp = self.store.similarity_search(query, k=k)
        logger.debug(rsp)
        if expand_cols:
            return str(sep.join([f"{x.page_content}: {x.metadata}" for x in rsp]))
        else:
            return str(sep.join([f"{x.page_content}" for x in rsp]))

    def write(self):
        """根据用户给定的Document（JSON / XLSX等）文件，进行index与库的初始化"""
        raise NotImplementedError

    def add(self, texts: list[str], *args, **kwargs) -> list[str]:
        """FIXME: 目前add之后没有更新store"""
        return self.store.add_texts(texts)

    def delete(self, *args, **kwargs):
        """目前langchain没有提供del接口"""
        raise NotImplementedError


if __name__ == '__main__':
    faiss_store = FaissStore(Path('qcs/qcs_4w.json'))
    logger.info(faiss_store.search('油皮洗面奶'))
    faiss_store.add([f'油皮洗面奶-{i}' for i in range(3)])
    logger.info(faiss_store.search('油皮洗面奶'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/4/29 16:07
@Author  : alexanderwu
@File    : common.py
"""
import ast
import inspect
import os
import re
import asyncio
import uuid
from typing import List, Tuple

from schema_agents.logs import logger
from contextvars import ContextVar

current_session = ContextVar('current_session')

def check_cmd_exists(command) -> int:
    """ 检查命令是否存在
    :param command: 待检查的命令
    :return: 如果命令存在，返回0，如果不存在，返回非0
    """
    check_command = 'command -v ' + command + ' >/dev/null 2>&1 || { echo >&2 "no mermaid"; exit 1; }'
    result = os.system(check_command)
    return result


class OutputParser:

    @classmethod
    def parse_blocks(cls, text: str):
        # 首先根据"##"将文本分割成不同的block
        blocks = text.split("##")

        # 创建一个字典，用于存储每个block的标题和内容
        block_dict = {}

        # 遍历所有的block
        for block in blocks:
            # 如果block不为空，则继续处理
            if block.strip() != "":
                # 将block的标题和内容分开，并分别去掉前后的空白字符
                block_title, block_content = block.split("\n", 1)
                # LLM可能出错，在这里做一下修正
                if block_title[-1] == ":":
                    block_title = block_title[:-1]
                block_dict[block_title.strip()] = block_content.strip()

        return block_dict

    @classmethod
    def parse_code(cls, text: str, lang: str = "") -> str:
        pattern = rf'```{lang}.*?\s+(.*?)```'
        match = re.search(pattern, text, re.DOTALL)
        if match:
            code = match.group(1)
        else:
            raise Exception
        return code

    @classmethod
    def parse_str(cls, text: str):
        text = text.split("=")[-1]
        text = text.strip().strip("'").strip("\"")
        return text

    @classmethod
    def parse_file_list(cls, text: str) -> list[str]:
        # Regular expression pattern to find the tasks list.
        pattern = r'\s*(.*=.*)?(\[.*\])'

        # Extract tasks list string using regex.
        match = re.search(pattern, text, re.DOTALL)
        if match:
            tasks_list_str = match.group(2)

            # Convert string representation of list to a Python list using ast.literal_eval.
            tasks = ast.literal_eval(tasks_list_str)
        else:
            tasks = text.split("\n")
        return tasks

    @classmethod
    def parse_data(cls, data):
        block_dict = cls.parse_blocks(data)
        parsed_data = {}
        for block, content in block_dict.items():
            # 尝试去除code标记
            try:
                content = cls.parse_code(text=content)
            except Exception:
                pass

            # 尝试解析list
            try:
                content = cls.parse_file_list(text=content)
            except Exception:
                pass
            parsed_data[block] = content
        return parsed_data

    @classmethod
    def parse_data_with_mapping(cls, data, mapping):
        block_dict = cls.parse_blocks(data)
        parsed_data = {}
        for block, content in block_dict.items():
            # 尝试去除code标记
            try:
                content = cls.parse_code(text=content)
            except Exception:
                pass
            typing_define = mapping.get(block, None)
            if isinstance(typing_define, tuple):
                typing = typing_define[0]
            else:
                typing = typing_define
            if typing == List[str] or typing == List[Tuple[str, str]]:
                # 尝试解析list
                try:
                    content = cls.parse_file_list(text=content)
                except Exception:
                    pass
            # TODO: 多余的引号去除有风险，后期再解决
            # elif typing == str:
            #     # 尝试去除多余的引号
            #     try:
            #         content = cls.parse_str(text=content)
            #     except Exception:
            #         pass
            parsed_data[block] = content
        return parsed_data


class CodeParser:

    @classmethod
    def parse_block(cls, block: str, text: str) -> str:
        blocks = cls.parse_blocks(text)
        for k, v in blocks.items():
            if block in k:
                return v
        return ""

    @classmethod
    def parse_blocks(cls, text: str):
        # 首先根据"##"将文本分割成不同的block
        blocks = text.split("##")

        # 创建一个字典，用于存储每个block的标题和内容
        block_dict = {}

        # 遍历所有的block
        for block in blocks:
            # 如果block不为空，则继续处理
            if block.strip() != "":
                # 将block的标题和内容分开，并分别去掉前后的空白字符
                block_title, block_content = block.split("\n", 1)
                block_dict[block_title.strip()] = block_content.strip()

        return block_dict

    @classmethod
    def parse_code(cls, block: str, text: str, lang: str = "") -> str:
        if block:
            text = cls.parse_block(block, text)
        pattern = rf'```{lang}.*?\s+(.*?)```'
        match = re.search(pattern, text, re.DOTALL)
        if match:
            code = match.group(1)
        else:
            logger.error(f"{pattern} not match following text:")
            logger.error(text)
            raise Exception
        return code

    @classmethod
    def parse_str(cls, block: str, text: str, lang: str = ""):
        code = cls.parse_code(block, text, lang)
        code = code.split("=")[-1]
        code = code.strip().strip("'").strip("\"")
        return code

    @classmethod
    def parse_file_list(cls, block: str, text: str, lang: str = "") -> list[str]:
        # Regular expression pattern to find the tasks list.
        code = cls.parse_code(block, text, lang)
        print(code)
        pattern = r'\s*(.*=.*)?(\[.*\])'

        # Extract tasks list string using regex.
        match = re.search(pattern, code, re.DOTALL)
        if match:
            tasks_list_str = match.group(2)

            # Convert string representation of list to a Python list using ast.literal_eval.
            tasks = ast.literal_eval(tasks_list_str)
        else:
            raise Exception
        return tasks


class NoMoneyException(Exception):
    """Raised when the operation cannot be completed due to insufficient funds"""

    def __init__(self, amount, message="Insufficient funds"):
        self.amount = amount
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message} -> Amount required: {self.amount}'


def print_members(module, indent=0):
    """
    https://stackoverflow.com/questions/1796180/how-can-i-get-a-list-of-all-classes-within-current-module-in-python
    :param module:
    :param indent:
    :return:
    """
    prefix = ' ' * indent
    for name, obj in inspect.getmembers(module):
        print(name, obj)
        if inspect.isclass(obj):
            print(f'{prefix}Class: {name}')
            # print the methods within the class
            if name in ['__class__', '__base__']:
                continue
            print_members(obj, indent + 2)
        elif inspect.isfunction(obj):
            print(f'{prefix}Function: {name}')
        elif inspect.ismethod(obj):
            print(f'{prefix}Method: {name}')

class EventBus:
    """An event bus class."""

    def __init__(self, name, logger=None):
        """Initialize the event bus."""
        self._callbacks = {}
        self._logger = logger
        self.name = name

    def on(self, event_name, func):
        """Register an event callback."""
        if self._callbacks.get(event_name):
            self._callbacks[event_name].add(func)
        else:
            self._callbacks[event_name] = {func}
        return func

    def once(self, event_name, func):
        """Register an event callback that only run once."""
        if self._callbacks.get(event_name):
            self._callbacks[event_name].add(func)
        else:
            self._callbacks[event_name] = {func}
        # mark once callback
        self._callbacks[event_name].once = True
        return func

    async def aemit(self, event_name, *data):
        """Trigger an event."""
        futures = []
        for func in self._callbacks.get(event_name, []):
            try:
                if inspect.iscoroutinefunction(func):
                    futures.append(func(*data))
                else:
                    func(*data)
                if hasattr(func, "once"):
                    self.off(event_name, func)
            except Exception as e:
                if self._logger:
                    self._logger.error(
                        "Error in event callback: %s, %s, error: %s",
                        event_name,
                        func,
                        e,
                    )
        await asyncio.gather(*futures)

    def emit(self, event_name, *data):
        """Trigger an event."""
        for func in self._callbacks.get(event_name, []):
            try:
                if inspect.iscoroutinefunction(func):
                    asyncio.get_running_loop().create_task(func(*data))
                else:
                    func(*data)
                if hasattr(func, "once"):
                    self.off(event_name, func)
            except Exception as e:
                if self._logger:
                    self._logger.error(
                        "Error in event callback: %s, %s, error: %s",
                        event_name,
                        func,
                        e,
                    )

    def off(self, event_name, func=None):
        """Remove an event callback."""
        if not func:
            del self._callbacks[event_name]
        else:
            self._callbacks.get(event_name, []).remove(func)

    async def stream_callback(self, message):
        if message["type"] == "function_call":
            if message["status"] == "in_progress":
                print(message["arguments"], end="")
            else:
                print(f'\nGenerating {message["name"]} ({message["status"]}): {message["arguments"]}')

    def register_default_events(self):
        self.on("stream", self.stream_callback)


import pytest
import os
from schema_agents.tools.code_interpreter import CodeInterpreter, extract_stdout, extract_stderr

class TestCodeInterpreter:
    @pytest.fixture(autouse=True)
    def init_code_interpreter(self):
        self.work_dir_root = "./.data"
        os.makedirs(self.work_dir_root, exist_ok=True)
        self.code_interpreter = CodeInterpreter(work_dir_root=self.work_dir_root)

    def test_execute_code(self):
        result = self.code_interpreter.execute_code("print('hello world')")
        stdout = extract_stdout(result)
        assert result["status"] == "ok", f"expected 'ok', got {result['status']}"
        assert stdout == "hello world\n", f"expected 'hello world\\n', got {stdout}"

        result = self.code_interpreter.execute_code("import numpy as np; print(np.random.rand(3))")
        assert result["status"] == "ok", f"expected 'ok', got {result['status']}"

        result = self.code_interpreter.execute_code("x = 1/0") # to induce ZeroDivisionError
        assert result["status"] == "error", f"expected 'error', got {result['status']}"
        assert "ZeroDivisionError" in result["traceback"], f"expected 'ZeroDivisionError' in traceback"

    def test_execute_query(self):
        output = self.code_interpreter.execute_query("say 'hello world'")
        assert "hello world" in output.lower(), f"expected 'hello world' in output, got {output}"

        output = self.code_interpreter.execute_query("create a random array and make sure the shape is (3,3)")
        assert "(3, 3)" in output, f"expected '(3, 3)' in output, got {output}"

        output = self.code_interpreter.execute_query("x = 1/0")  # to induce division by zero
        # The model will use try and except to capture the error, and print the error message.
        assert "division by zero" in output, f"expected 'division by zero' in output, got {output}"
        
        output = self.code_interpreter.execute_query("make a simple plot")
        assert "successfully executed" in output, f"expected 'successfully executed' in output, got {output}"
        output = self.code_interpreter.execute_query("save the plot into a png file named 'plot.png'")
        assert "successfully executed" in output, f"expected 'successfully executed' in output, got {output}"
        assert os.path.exists(os.path.join(self.code_interpreter.work_dir, "plot.png")), f"expected 'plot.png' in {self.code_interpreter.work_dir}, got {os.listdir(self.code_interpreter.work_dir)}"
        

    def test_tear_down(self):
        self.code_interpreter.tearDown()
        assert self.code_interpreter.km is None, "Kernel Manager is not shut down correctly."
        assert self.code_interpreter.kc is None, "Kernel Client is not shut down correctly."



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/20 12:15
@Author  : alexanderwu
@File    : memory.py
"""
from collections import defaultdict, OrderedDict
from typing import Iterable, Callable
from pydantic import BaseModel

from schema_agents.schema import Message


class Memory:
    """The most basic memory: super-memory"""

    def __init__(self):
        """Initialize an empty storage list and an empty index dictionary"""
        self.storage: list[Message] = []
        self.index: OrderedDict[Callable, list[Message]] = OrderedDict()

    def add(self, message: Message):
        """Add a new message to storage, while updating the index"""
        if message in self.storage:
            return
        self.storage.append(message)
        if message.cause_by:
            if message.cause_by not in self.index:
                self.index[message.cause_by] = []
            self.index[message.cause_by].append(message)


    def add_batch(self, messages: Iterable[Message]):
        for message in messages:
            self.add(message)

    def get_by_role(self, role: str) -> list[Message]:
        """Return all messages of a specified role"""
        return [message for message in self.storage if message.role == role]

    def get_by_content(self, content: str) -> list[Message]:
        """Return all messages containing a specified content"""
        return [message for message in self.storage if content in message.content]
    
    def get_by_schemas(self, schemas: list[BaseModel]) -> list[Message]:
        ret = []
        for schema in schemas:
            if schema is str:
                ret += [message for message in self.storage if not message.data and isinstance(message.content, str)]
            else:
                ret += [message for message in self.storage if message.data and isinstance(message.data, schema)]
        return ret

    def delete(self, message: Message):
        """Delete the specified message from storage, while updating the index"""
        self.storage.remove(message)
        if message.cause_by and message in self.index[message.cause_by]:
            self.index[message.cause_by].remove(message)

    def clear(self):
        """Clear storage and index"""
        self.storage = []
        self.index = defaultdict(list)

    def count(self) -> int:
        """Return the number of messages in storage"""
        return len(self.storage)

    def try_remember(self, keyword: str) -> list[Message]:
        """Try to recall all messages containing a specified keyword"""
        return [message for message in self.storage if keyword in message.content]

    def get(self, k=0) -> list[Message]:
        """Return the most recent k memories, return all when k=0"""
        return self.storage[-k:]

    def remember(self, observed: list[Message], k=10) -> list[Message]:
        """remember the most recent k memories from observed Messages, return all when k=0"""
        already_observed = self.get(k)
        news: list[Message] = []
        for i in observed:
            if i in already_observed:
                continue
            news.append(i)
        return news

    def get_by_action(self, action: Callable) -> list[Message]:
        """Return all messages triggered by a specified Action"""
        return self.index[action]

    def get_by_actions(self, actions: Iterable[Callable]) -> list[Message]:
        """Return all messages triggered by specified Actions"""
        rsp = []
        for key in self.index.keys():
            if key in actions:
                rsp += self.index[key]
        return rsp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/5 23:00
@Author  : alexanderwu
@File    : base_chatbot.py
"""
from abc import ABC, abstractmethod
from dataclasses import dataclass


@dataclass
class BaseChatbot(ABC):
    """Abstract GPT class"""
    mode: str = "API"

    @abstractmethod
    def ask(self, msg: str) -> str:
        """Ask GPT a question and get an answer"""

    @abstractmethod
    def ask_batch(self, msgs: list) -> str:
        """Ask GPT multiple questions and get a series of answers"""

    @abstractmethod
    def ask_code(self, msgs: list) -> str:
        """Ask GPT multiple questions and get a piece of code"""




from pydantic import BaseModel, Field
from typing import Optional
import asyncio
from functools import partial

from schema_agents.schema import Message
from schema_agents.role import Role
from .project_manager import SoftwareRequirement
from .data_engineer import create_data_engineer
from .web_developer import create_web_developer, ReactUI
from schema_agents.tools.code_interpreter import create_mock_client
from schema_agents.teams import Team

import yaml

class GenerateUserForm(BaseModel):
    """Based on user query, create a JSON Schema Form Dialog using react-jsonschema-form to get more information about what the user want to create.
    The aim is to gather the information needed to create the software requirement document.
    Whenever possible, try to propose the options for the user to choose from, instead of asking the user to type in the text."""
    form_schema: str = Field(description="json schema for the fields, in yaml format")
    ui_schema: Optional[str] = Field(None, description="customized ui schema for rendering the form, json string, no need to escape quotes, in yaml format")
    submit_label: Optional[str] = Field("Submit", description="Submit button label")

class UserRequirements(BaseModel):
    """User requirements for the software."""
    goal: str = Field(description="The goal of the user.")
    ui: str = Field(description="Requirements for the user interface.")
    data: str = Field(description="Requirements for the data.")
    validation: str = Field(description="Additional information for testing, e.g. test data and the expected outcome")
    notes: str = Field(description="Additional notes.")
    
class UserClarification(BaseModel):
    """User submitted form data."""
    form_data: str = Field(description="Form data in json format")

class GetExtraInformation(BaseModel):
    """Extra information needed to be able to work on the task."""
    content: str = Field(description="The information.")
    summary: str = Field(description="Summary of what you already get.")

async def show_message(client, message: Message):
    """Show message to the user."""
    await client.set_output(message.content)

async def clarify_user_request(client, user_query: str, role: Role) -> UserClarification:
    """Clarify user request by prompting to the user with a form."""
    config = await role.aask(user_query, GenerateUserForm)
    fm = await client.show_dialog(
        src="https://oeway.github.io/imjoy-json-schema-form/",
        config={
            "schema": config.form_schema and yaml.safe_load(config.form_schema),
            "ui_schema": config.ui_schema and yaml.safe_load(config.ui_schema),
            "submit_label": config.submit_label,
        }
    )
    form = await fm.get_data()
    return UserClarification(form_data=str(form['formData']))

async def create_user_requirements(req: UserClarification, role: Role) -> UserRequirements:
    """Create user requirement."""
    return await role.aask(req, UserRequirements)

async def create_software_requirements(req: UserRequirements, role: Role) -> SoftwareRequirement:
    """Create software requirement."""
    return await role.aask(req, SoftwareRequirement)

async def deploy_app(ui: ReactUI, role: Role):
    """Deploy the app for sharing."""
    # serve_plugin(ui)
    print("Deploying the app...")


def create_hpa_database_explorer(client, investment):
    """recruit roles to cooperate"""
    team = Team(name="HPA Database Explorer", profile="A team of roles to create software for image analysis.", goal="Create a software for image analysis.", investment=investment)
    ux_manager = Role(name="Luisa",
        profile="UX Manager",
        goal="Focus on understanding the user's needs and experience. Understand the user needs by interacting with user and communicate these findings to the project manager by calling `UserRequirements`.",
        constraints=None,
        actions=[partial(clarify_user_request, client), create_user_requirements])

    project_manager = Role(name="Alice",
                profile="Project Manager",
                goal="Efficiently communicate with the user and translate the user's needs into software requirements",
                constraints=None,
                actions=[create_software_requirements])

    web_developer  = create_web_developer(client=client)
    data_engineer = create_data_engineer(client=client)
    devops = Role(name="Bruce",
                profile="DevOps",
                goal="Deploy the software to the cloud and make it available to the user.",
                constraints=None,
                actions=[deploy_app])  

    team.hire([ux_manager, project_manager, web_developer, data_engineer, web_developer, devops])
    return team

async def test():
    lab = create_hpa_database_explorer(client=create_mock_client())
    await lab.handle("create a cell counting software")

if __name__ == "__main__":
    asyncio.run(test())


import aiohttp
import asyncio
from aiohttp_sse_client import client as sse_client
import json
import uuid
import os
from jhub_client.api import JupyterKernelAPI, JupyterAPI

# Example code for jhub_client: https://github.com/Quansight/jhub-client/tree/master/tests

class KernelAPI(JupyterKernelAPI):
    async def execute_code(self, username, code, wait=True, timeout=None):
        msg_id = str(uuid.uuid4())

        await self.websocket.send_json(
            self.request_execute_code(msg_id, username, code)
        )

        if not wait:
            return None

        async for msg_text in self.websocket:
            if msg_text.type != aiohttp.WSMsgType.TEXT:
                return False

            # TODO: timeout is ignored

            msg = msg_text.json()

            if "parent_header" in msg and msg["parent_header"].get("msg_id") == msg_id:
                # These are responses to our request
                if msg["channel"] == "iopub":
                    if msg["msg_type"] == "execute_result":
                        return msg["content"]["data"]["text/plain"]
                    elif msg["msg_type"] == "stream":
                        return msg["content"]["text"]
                    # cell did not produce output
                    elif msg["content"].get("execution_state") == "idle":
                        return ""

class BinderServer():
    def __init__(self, repo='imjoy-team/imjoy-binder-image', branch='master', config=None):
        self.repo = repo
        self.branch = branch
        self.jhub_config = config

    async def start(self, load_existing=True):
        if load_existing:
            if os.path.exists('.jhub_config.json'):
                try:
                    with open('.jhub_config.json', 'r') as f:
                        jhub_config = json.load(f)
                    jupyter = JupyterAPI(jhub_config['url'], jhub_config['token'])
                    async with jupyter:
                        specs = await jupyter.list_kernel_specs()
                        if len(specs) > 0:
                            self.jhub_config = jhub_config
                            print("Existing Binder server found.")
                            return self.jhub_config
                except Exception as e:
                    print("Failed to reuse existing Binder server config.")

        binder_api_url = f"https://notebooks.gesis.org/binder/build/gh/{self.repo}/{self.branch}"
        # Start Binder build
        async with aiohttp.ClientSession() as session:
            async with session.post(binder_api_url) as response:
                if response.status == 200:
                    print("Binder build starting...")
                else:
                    print(f"Failed to start Binder build: {response.status}")
        
        # Listen to SSE events for build status
        async with sse_client.EventSource(binder_api_url) as event_source:
            jhub_config = None
            async for event in event_source:
                data = json.loads(event.data)
                if data['phase'] == 'ready':
                    # example: MessageEvent(type=None, message='', data='{"phase": "ready", "message": "server running at https://notebooks.gesis.org/binder/jupyter/user/imjoy-team-imjoy-binder-image-9fc2zw8r/\\n", "image": "gesiscss/binder-r2d-g5b5b759-imjoy-2dteam-2dimjoy-2dbinder-2dimage-560b33:83282e916f96e5cc28e76775965c1601331ace18", "repo_url": "https://github.com/imjoy-team/imjoy-binder-image", "token": "vXFZ0kV6S3auR2nbUCYXbA", "binder_ref_url": "https://github.com/imjoy-team/imjoy-binder-image/tree/83282e916f96e5cc28e76775965c1601331ace18", "binder_launch_host": "https://mybinder.org/", "binder_request": "v2/gh/imjoy-team/imjoy-binder-image/master", "binder_persistent_request": "v2/gh/imjoy-team/imjoy-binder-image/83282e916f96e5cc28e76775965c1601331ace18", "url": "https://notebooks.gesis.org/binder/jupyter/user/imjoy-team-imjoy-binder-image-9fc2zw8r/"}', origin='https://notebooks.gesis.org', last_event_id='')
                    # Usage: jupyter = JupyterAPI(hub.hub_url / "user" / username, hub.api_token)
                    self.username = self.jhub_config['url'].split('/')[-2]
                    jhub_config = data
                    print("Binder build finished.", data['message'])
                    break
                if data['phase'] == 'failed':
                    print("Binder build failed.", data['message'])
                    break
                else:
                    print(f"Binder {data['phase']}: {data['message']}")
        if jhub_config is None:
            raise Exception(f"Failed to start Binder server: {data['message']}")
        with open('.jhub_config.json', 'w') as f:
            json.dump(self.jhub_config, f)
        return self.jhub_config

        
    async def start_kernel(self):
        jupyter = JupyterAPI(self.jhub_config['url'], self.jhub_config['token'])
        async with jupyter:
            kernel_id = (await jupyter.create_kernel())["id"]
            kernel = KernelAPI(
                jupyter.api_url / "kernels" / kernel_id, jupyter.api_token
            )
            return kernel
    
    async def stop_kernel(self, kernel_id):
        jupyter = JupyterAPI(self.jhub_config['url'], self.jhub_config['token'])
        async with jupyter:
            await jupyter.delete_kernel(kernel_id)

    async def execute_code(self, kernel, code):
        async with kernel:
            return await kernel.execute_code(
                self.username,
                code
            )

async def test_binder():
    binder = BinderServer()
    await binder.start()
    kernel = await binder.start_kernel()
    print(await binder.execute_code(kernel, "print('hello world')"))
    await binder.stop_kernel(kernel.kernel_id)
    
if __name__ == '__main__':
    asyncio.run(test_binder())



import pytest
import os
import time
from schema_agents.hyphaichat import HyphaAIChat
from schema_agents.tools.code_interpreter import CodeInterpreter
from pydantic import BaseModel
from simpleaichat.utils import fd

params = {"temperature": 0.0, "max_tokens": 1000}

class Author(BaseModel):
    name: str = fd("The name of the author.")
    email: str = fd("The email of the author.")

class Story(BaseModel):
    """Story."""
    name: str = fd("The name of the story.")
    year: int = fd("The year of the story.")
    text: str = fd("If selected, arguments to pass to the tool.")
    pages: int = fd("The number of pages in the story.")
    author: Author = fd("The author of the story.")

def write_file(fancy_name: str, content: str):
    """Write content to file."""
    print("=======> Writing to file: ", fancy_name)
    with open(fancy_name, "w") as f:
        f.write(content)
    # print("Error occurred while writing file: " + file_name)


def read_file(file_name: str):
    """Read content from file."""
    print("=======> Reading from file: ", file_name)
    with open(file_name, "r") as f:
        content = f.read()
    
    return content

class TestCodeInterpreter:
    @pytest.fixture(autouse=True)
    def init_code_interpreter(self):
        self.work_dir_root = "./.data"
        os.makedirs(self.work_dir_root, exist_ok=True)
        self.code_interpreter = CodeInterpreter(work_dir_root=self.work_dir_root)
        self.bot = HyphaAIChat(params=params, console=True, system="""You are Story-GPT, an AI designed to autonomously write stories.""")
        
    def test_functions(self):
        initial_message_length = len(self.bot.get_session().messages)
        response = self.bot("Write a short story in 100 words about yourself and save it into a file.", functions=[write_file, read_file])
        assert response.get('function') == 'write_file'
        assert len(self.bot.get_session().messages) == initial_message_length + 3
        
        response = self.bot("Read the story and generate a short title for it.", functions=[write_file, read_file])
        assert response.get('function') == 'read_file'
        assert len(self.bot.get_session().messages) == initial_message_length + 6

        response = self.bot("Execute Python script to count the #words in the story file", functions=[write_file, read_file, self.code_interpreter.execute_code])
        assert response.get('function') == 'execute_code'
        assert len(self.bot.get_session().messages) == initial_message_length + 9
        
        response = self.bot("Tell me a word which is interesting to learn", functions=[write_file, read_file])
        assert response.get('function') == None
        assert len(self.bot.get_session().messages) == initial_message_length + 11
        
        
    def test_io_schema(self):
        initial_message_length = len(self.bot.get_session().messages)
        response = self.bot("write a short story in 100 words about yourself and save it into a file named 'my_story.txt'.", functions=[write_file, read_file], output_schema=Story)
        assert response.get('function') == 'write_file'
        story = response.get('response')
        Story.model_validate(story)
        # If output_schema is specified, the length of the messages won't increase
        assert len(self.bot.get_session().messages) == initial_message_length + 3
        



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/7/4 10:53
@Author  : alexanderwu
@File    : mermaid.py
"""
import os
import subprocess
from pathlib import Path

from schema_agents.const import PROJECT_ROOT
from schema_agents.logs import logger
from schema_agents.utils.common import check_cmd_exists

IS_DOCKER = os.environ.get('AM_I_IN_A_DOCKER_CONTAINER', 'false').lower()


def mermaid_to_file(mermaid_code, output_file_without_suffix, width=2048, height=2048) -> int:
    """suffix: png/svg/pdf

    :param mermaid_code: mermaid code
    :param output_file_without_suffix: output filename
    :param width:
    :param height:
    :return: 0 if succed, -1 if failed
    """
    # Write the Mermaid code to a temporary file
    tmp = Path(f'{output_file_without_suffix}.mmd')
    tmp.write_text(mermaid_code, encoding='utf-8')

    if check_cmd_exists('mmdc') != 0:
        logger.warning(
            "RUN `npm install -g @mermaid-js/mermaid-cli` to install mmdc")
        return -1

    for suffix in ['pdf', 'svg', 'png']:
        output_file = f'{output_file_without_suffix}.{suffix}'
        # Call the `mmdc` command to convert the Mermaid code to a PNG
        logger.info(f"Generating {output_file}..")
        if IS_DOCKER == 'true':
            subprocess.run(['mmdc', '-p', '/app/schema_agents/puppeteer-config.json', '-i',
                           str(tmp), '-o', output_file, '-w', str(width), '-H', str(height)])
        else:
            subprocess.run(['mmdc', '-i', str(tmp), '-o',
                           output_file, '-w', str(width), '-H', str(height)])
    return 0


MMC1 = """classDiagram
    class Main {
        -SearchEngine search_engine
        +main() str
    }
    class SearchEngine {
        -Index index
        -Ranking ranking
        -Summary summary
        +search(query: str) str
    }
    class Index {
        -KnowledgeBase knowledge_base
        +create_index(data: dict)
        +query_index(query: str) list
    }
    class Ranking {
        +rank_results(results: list) list
    }
    class Summary {
        +summarize_results(results: list) str
    }
    class KnowledgeBase {
        +update(data: dict)
        +fetch_data(query: str) dict
    }
    Main --> SearchEngine
    SearchEngine --> Index
    SearchEngine --> Ranking
    SearchEngine --> Summary
    Index --> KnowledgeBase"""

MMC2 = """sequenceDiagram
    participant M as Main
    participant SE as SearchEngine
    participant I as Index
    participant R as Ranking
    participant S as Summary
    participant KB as KnowledgeBase
    M->>SE: search(query)
    SE->>I: query_index(query)
    I->>KB: fetch_data(query)
    KB-->>I: return data
    I-->>SE: return results
    SE->>R: rank_results(results)
    R-->>SE: return ranked_results
    SE->>S: summarize_results(ranked_results)
    S-->>SE: return summary
    SE-->>M: return summary"""


if __name__ == '__main__':
    # logger.info(print_members(print_members))
    mermaid_to_file(MMC1, '/tmp/1.png')
    mermaid_to_file(MMC2, '/tmp/2.png')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/1 11:59
@Author  : alexanderwu
@File    : const.py
"""
import os
from pathlib import Path

PROJECT_ROOT = Path(os.environ.get('PROJECT_ROOT', os.getcwd()))
DATA_PATH = PROJECT_ROOT / 'data'

MEM_TTL = 24 * 30 * 3600



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import asyncio
import inspect
import json
import uuid
import traceback
import types
import typing
from functools import partial
from inspect import signature
from typing import Iterable, Optional, Union, Callable
from pydantic import BaseModel

from schema_agents.logs import logger
from schema_agents.utils import (parse_special_json,
                              schema_to_function)
from schema_agents.llm import LLM
from schema_agents.schema import Message
from schema_agents.memory.long_term_memory import LongTermMemory
from schema_agents.utils.common import EventBus
from pydantic import BaseModel
from schema_agents.utils.common import current_session
from contextlib import asynccontextmanager
from contextvars import copy_context

PREFIX_TEMPLATE = """You are a {profile}, named {name}, your goal is {goal}, and the constraint is {constraints}. """

@asynccontextmanager
async def create_session_context(session_id):
    current_session.set(session_id)
    yield copy_context()

class RoleSetting(BaseModel):
    """Role setting"""
    name: str
    profile: str
    goal: str
    constraints: Optional[str]
    desc: str

    def __str__(self):
        return f"{self.name}({self.profile})"

    def __repr__(self):
        return self.__str__()


class Role:
    """Role is a person or group who has a specific job or purpose within an organization."""
    def __init__(self, name="", profile="", goal="", constraints=None, desc="", long_term_memory: Optional[LongTermMemory]=None, event_bus:EventBus =None, actions: list[Callable] = None, model: str=None):
        self._llm = LLM(model=model)
        self._setting = RoleSetting(name=name, profile=profile, goal=goal, constraints=constraints, desc=desc)
        self._states = []
        self._actions = actions or []
        self._role_id = str(self._setting)
        self._input_schemas = []
        self._output_schemas = []
        self._action_index = {}
        self._user_support_actions = []
        self._watch_actions = set()
        self.long_term_memory = long_term_memory
        if event_bus:
            self.set_event_bus(event_bus)
        else:
            self.set_event_bus(EventBus(f"{self._setting.profile} - {self._setting.name}"))
        self._init_actions(self._actions)

    def _reset(self):
        self._states = []
        self._actions = []


    def _watch(self, actions: Iterable[Union[str, BaseModel]]):
        """Watch actions."""
        self._watch_actions.update(actions)
    
    def set_event_bus(self, event_bus: EventBus):
        """Set event bus."""
        self._event_bus = event_bus

        async def handle_message(msg):
            if msg.data and type(msg.data) in self._watch_actions:
                await self.handle(msg)
            elif msg.data is None and str in self._watch_actions:
                await self.handle(msg)

        self._event_bus.on("message", handle_message)
        logger.info(f"Mounting {self._setting} to event bus: {self._event_bus.name}.")
    
    def get_event_bus(self):
        """Get event bus."""
        return self._event_bus

    @property
    def profile(self):
        """Get profile."""
        return self._setting.profile

    def _get_prefix(self):
        """Get prefix."""
        if self._setting.desc:
            return self._setting.desc
        return PREFIX_TEMPLATE.format(**self._setting.dict())
    
    @property
    def user_support_actions(self):
        return self._user_support_actions
    
    @property
    def prefix(self):
        return self._get_prefix()

    def _init_actions(self, actions):
        self._output_schemas = []
        self._input_schemas = []
        for action in actions:
            if isinstance(action, partial):
                action.__doc__ = action.func.__doc__
                action.__name__ = action.func.__name__
            assert action.__doc__, "Action must have docstring"
            assert isinstance(action, (partial, types.FunctionType, types.MethodType)), f"Action must be function, but got {action}"
            sig = signature(action)
            positional_annotation = [p.annotation for p in sig.parameters.values() if p.kind == p.POSITIONAL_OR_KEYWORD][0]
            assert positional_annotation == str or isinstance(positional_annotation, typing._UnionGenericAlias) or issubclass(positional_annotation, BaseModel), f"Action only support pydantic BaseModel, typing.Union or str, but got {positional_annotation}"
            output_schemas = [sig.return_annotation] if not isinstance(sig.return_annotation, typing._UnionGenericAlias) else list(sig.return_annotation.__args__)
            input_schemas = [positional_annotation] if not isinstance(positional_annotation, typing._UnionGenericAlias) else list(positional_annotation.__args__)
            self._output_schemas += output_schemas
            self._input_schemas += input_schemas
            for schema in input_schemas:
                if schema not in self._action_index:
                    self._action_index[schema] = [action]
                else:
                    self._action_index[schema].append(action)
            # mark as user support action if the input schema is str
            if str in input_schemas:
                self._user_support_actions.append(action)
        self._output_schemas = list(set(self._output_schemas))
        self._input_schemas = list(set(self._input_schemas))
        self._watch(self._input_schemas)
        
        self._reset()
        for idx, action in enumerate(actions):
            self._actions.append(action)
            self._states.append(f"{idx}. {action}")
    
    async def _run_action(self, action, msg):
        sig = signature(action)
        keys = [p.name for p in sig.parameters.values() if p.kind == p.POSITIONAL_OR_KEYWORD and p.annotation == Role]
        kwargs = {k: self for k in keys}
        pos = [p for p in sig.parameters.values() if p.kind == p.POSITIONAL_OR_KEYWORD and p.annotation != Role]
        for p in pos:
            if not msg.data and isinstance(msg.content, str):
                kwargs[p.name] = msg.content
                msg.processed_by.add(self)
                break
            elif msg.data and isinstance(msg.data, p.annotation.__args__ if isinstance(p.annotation, typing._UnionGenericAlias) else p.annotation):
                kwargs[p.name] = msg.data
                msg.processed_by.add(self)
                break
            if p.name not in kwargs:
                kwargs[p.name] = None
        
        if inspect.iscoroutinefunction(action):
            return await action(**kwargs)
        else:
            return action(**kwargs)

    def can_handle(self, message: Message) -> bool:
        """Check if the role can handle the message."""
        context_class = message.data.__class__ if message.data else type(message.content)
        if context_class in self._input_schemas:
            return True
        return False
    
    # @retry(stop=stop_after_attempt(2), wait=wait_fixed(1))
    async def handle(self, msg: Union[str, Message], session_id=None) -> list[Message]:
        """Handle message"""
        if isinstance(msg, str):
            msg = Message(role="User", content=msg)
        if not self.can_handle(msg):
            raise ValueError(f"Invalid message, the role {self._setting} cannot handle the message: {msg}")
        _session_id = str(uuid.uuid4())
        msg.session_ids.append(_session_id)
        messages = []
        def on_message(new_msg):
            if _session_id in new_msg.session_ids:
                messages.append(new_msg)

        self._event_bus.on("message", on_message)
        try:
            context_class = msg.data.__class__ if msg.data else type(msg.content)
            responses = []
            if context_class in self._input_schemas:
                actions = self._action_index[context_class]
                for action in actions:
                    responses.append(self._run_action(action, msg))
            if session_id:
                async with create_session_context(session_id):
                    responses = await asyncio.gather(*responses)
            else:
                responses = await asyncio.gather(*responses)
            outputs = []  
            for response in responses:
                if not response:
                    continue
                # logger.info(response)
                if isinstance(response, str):
                    output = Message(content=response, role=self.profile, cause_by=action, session_ids=msg.session_ids.copy())
                else:
                    assert isinstance(response, BaseModel), f"Action must return pydantic BaseModel, but got {response}"
                    output = Message(content=response.json(), data=response, session_ids=msg.session_ids.copy(),
                                role=self.profile, cause_by=action)
                # self._rc.memory.add(output)
                # logger.debug(f"{response}")
                outputs.append(output)
                if session_id:
                    async with create_session_context(session_id):
                        await self._event_bus.aemit("message", output)
                else:
                    await self._event_bus.aemit("message", output)
        finally:
            self._event_bus.off("message", on_message)
        return messages

    # @retry(stop=stop_after_attempt(2), wait=wait_fixed(1))
    async def aask(self, req, output_schema=None, prompt=None):
        output_schema = output_schema or str
        if isinstance(req, str):
            messages = [{"role": "user", "content": req}]
            input_schema = None
        elif isinstance(req, dict):
            messages = [req]
            input_schema = None
        elif isinstance(req, BaseModel):
            input_schema = req.__class__
            messages = [{"role": "function", "name": input_schema.__name__, "content": req.json()}]
        else:
            assert isinstance(req, list)
            messages = []
            for r in req:
                if isinstance(r, str):
                    messages.append({"role": "user", "content": r})
                    input_schema = None
                elif isinstance(r, dict):
                    messages.append(r)
                    input_schema = None
                elif isinstance(r, BaseModel):
                    input_schema = r.__class__
                    messages.append({"role": "function", "name": input_schema.__name__, "content": r.json()})
                else:
                    raise ValueError(f"Invalid request {r}")
        
        assert output_schema is str or isinstance(output_schema, typing._UnionGenericAlias) or issubclass(output_schema, BaseModel)
        
        if input_schema:
            prefix = f"Please generate a response based on the result of `{input_schema.__name__}`. "
        else:
            prefix = ""
        if output_schema is str:
            output_types = []
            prompt = prompt or f"{prefix}"
            messages.append({"role": "user", "content": f"{prompt}"})
        elif isinstance(output_schema, typing._UnionGenericAlias):
            output_types = list(output_schema.__args__)
            schema_names = ",".join([f"`{s.__name__}`" for s in output_types])
            prompt = prompt or f"{prefix}You MUST call one of the following functions: {schema_names}. DO NOT return text directly."
            messages.append({"role": "user", "content": f"{prompt}"})
        else:
            output_types = [output_schema]
            prompt = prompt or f"{prefix}You MUST call the `{output_schema.__name__}` function."
            messages.append({"role": "user", "content": f"{prompt}"})
        system_msgs = [self._get_prefix()]
        
        if output_schema is str:
            function_call = "none"
            return await self._llm.aask(messages, system_msgs, functions=[input_schema] if input_schema else [], function_call=function_call, event_bus=self._event_bus)

        functions = [schema_to_function(s) for s in set(output_types + ([input_schema] if input_schema else []))]
        if len(output_types) == 1:
            function_call = {"name": output_types[0].__name__}
        else:
            function_call = "auto"
        response = await self._llm.aask(messages, system_msgs, functions=functions, function_call=function_call, event_bus=self._event_bus)
        try:
            schema_names = ",".join([f"`{s.__name__}`" for s in output_types])
            assert not isinstance(response, str), f"Invalid response, you MUST call one of the following functions: {schema_names}. DO NOT return text directly."
            assert response["name"] in [s.__name__ for s in output_types], f"Invalid function name: {response['name']}"
            idx = [s.__name__ for s in output_types].index(response["name"])
            arguments = parse_special_json(response["arguments"])
            return output_types[idx].parse_obj(arguments)
        except Exception:
            logger.error(f"Failed to parse the response, error:\n{traceback.format_exc()}\nPlease regenerate to fix the error.")
            messages.append({"role": "assistant", "content": str(response)})
            messages.append({"role": "user", "content": f"Failed to parse the response, error:\n{traceback.format_exc()}\nPlease regenerate to fix the error."})
            response = await self._llm.aask(messages, system_msgs, functions=functions, function_call=function_call, event_bus=self._event_bus)
            assert response["name"] in [s.__name__ for s in output_types], f"Invalid function name: {response['name']}"
            idx = [s.__name__ for s in output_types].index(response["name"])
            arguments = json.loads(response["arguments"])
            return output_types[idx].parse_obj(arguments)
            
        